"""
Sample-parallel implementation of k-means clustering in PyTorch using MPI
"""
import time

import h5py
import torch
from mpi4py import MPI


class KMeans:
    """
    Sample-parallel k-means clustering in PyTorch with MPI.

    Each rank holds an exclusive chunk of the complete dataset (distributed along the sample axis) and determines
    locally which local sample belongs to which cluster before updating the centroids globally based on these local
    assignments and continuing with the next iteration.

    Sample parallelism for k-means requires communication at two points in the code, i.e.:
    1. Centroid initialization: Each rank determines a respective number of centroids from its local data chunk to be
       all-gathered by all ranks.
    2. Centroid update: Each rank needs to know all samples from other ranks in a certain cluster to update the
       centroids correctly, i.e., the samples assigned to a certain cluster on each local rank and their number need to
       be all-reduced to all ranks.

    Attributes
    ----------
    comm : MPI.Comm
        The MPI communicator.
    n_clusters : int
        The number of clusters.
    max_iter : int
        The maximum number of iterations.
    tol : float
        The convergence criterion.
    _inertia : float
        The inertia (quantity to be checked for convergence).
    _cluster_centers : Union[None, torch.Tensor]
        The centroids.
    _matching_centroids : torch.Tensor
        Indices of nearest centroid for each sample.
    """

    def __init__(
        self,
        comm: MPI.Comm = MPI.COMM_WORLD,
        n_clusters: int = 8,
        max_iter: int = 300,
        tol: float = -1.0,
    ) -> None:
        """
        Configure sample-parallel k-means clustering algorithm.

        Parameters
        ----------
        comm : MPI.Comm
            The MPI communicator.
        n_clusters : int
            The number of clusters, i.e., k.
        max_iter : int
            The maximum number of iterations to be performed.
        tol : float
            The tolerance for the convergence criterion.
        """
        self.comm = comm  # MPI communicator
        self.n_clusters = n_clusters  # Number of clusters
        self.max_iter = max_iter  # Maximum number of iterations
        self.tol = tol  # Tolerance for convergence criterion

        self._inertia = float("nan")
        self._cluster_centers = None
        self._matching_centroids = None

    def _initialize_centroids(self, x: torch.Tensor) -> None:
        """
        Randomly initialize the centroids.

        For sample parallelism, each rank holds an exclusive fraction of samples from the global dataset.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.
        """
        assert self.n_clusters >= self.comm.size  # Assume `n_clusters` >= `comm.size`.
        n_clusters_local = (
            self.n_clusters // self.comm.size
        )  # Determine number of local cluster centers per rank.
        # Choose first `n_cluster_local` randomly shuffled indices as centroid indices.
        # `torch.randperm(n)` returns random permutation of integers from 0 to n - 1.
        # `x.shape[0]` gives number of samples in data.
        x_local = x[torch.randperm(x.shape[0])[:n_clusters_local]]
        print(
            f"Rank {self.comm.rank}/{self.comm.size}: Local cluster centers have shape [#samples, #features] = "
            f"{list(x_local.shape)}."
        )
        # --- Communication required for sample-parallel: All-gather local centroids to all ranks ---
        self._cluster_centers = torch.empty(
            (self.n_clusters, x.shape[1]), dtype=torch.float
        )  # Initialize empty tensor of shape [# clusters, # features] on each rank to store all centroids.

        self.comm.Allgather(
            [x_local, MPI.FLOAT], [self._cluster_centers, MPI.FLOAT]
        )  # All-gather local centroids to all ranks.
        if self.comm.rank == 0:
            print(
                f"Global cluster centers have shape [#samples, #features] = {list(self._cluster_centers.shape)}.",
            )

    def _fit_to_cluster(self, x: torch.Tensor) -> torch.Tensor:
        """
        Determine the closest centroids for each sample in dataset as measured by their Euclidean distance.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to be clustered.

        Returns
        -------
        torch.Tensor
            Indices of matching centroids for each sample in dataset.
        """
        distances = torch.cdist(
            x, self._cluster_centers
        )  # Calculate Euclidean distances between samples and centroids.
        return distances.argmin(
            dim=1, keepdim=True
        )  # Determine index of nearest centroid for each sample.

    def fit(self, x: torch.Tensor) -> "KMeans":
        """
        Perform k-means clustering of given dataset.

        Parameters
        ----------
        x : torch.Tensor
            The dataset to cluster.

        Returns
        -------
        KMeans
            The fitted KMeans object containing final centroids.
        """
        self._initialize_centroids(x)  # Initialize centroids.
        new_cluster_centers = self._cluster_centers.clone()

        # Iteratively fit points to centroids.
        for idx in range(self.max_iter):
            # Determine nearest centroids.
            print(
                f"Rank {self.comm.rank}/{self.comm.size}: Iteration {idx}...",
            )
            self._matching_centroids = self._fit_to_cluster(
                x
            )  # Determine index of nearest centroid for each sample.

            # Update centroids.
            for i in range(self.n_clusters):
                # Locally determine samples in currently considered cluster i (binary encoding).
                selection_mask = (self._matching_centroids == i).type(torch.int64)

                # Locally accumulate samples and total number of samples in cluster i.
                assigned_points = (x * selection_mask).sum(
                    axis=0, keepdim=True
                )  # Local directed sum of samples in cluster i.

                # Local number of samples in cluster i.
                points_in_cluster = (
                    selection_mask.sum(axis=0, keepdim=True)
                    .clamp(0.0, torch.iinfo(torch.int64).max)
                    .type(torch.int64)
                )  # Clamp to 0 before communication

                # --- Communication required for sample-parallel: Allreduce local results ---
                assigned_points_global = torch.empty(
                    assigned_points.shape, dtype=torch.float
                )  # Initialize variable to store global directed sum of points assigned to cluster i.
                points_in_cluster_global = torch.empty(
                    points_in_cluster.shape, dtype=torch.int64
                )  # Initialize variable to store global number of points assigned to cluster i.
                self.comm.Allreduce(
                    assigned_points, assigned_points_global, op=MPI.SUM
                )  # All-reduce local directed sums of points assigned to cluster i.
                self.comm.Allreduce(
                    points_in_cluster, points_in_cluster_global, op=MPI.SUM
                )  # All-reduce local numbers of points assigned to cluster i.

                # Compute new centroids.
                new_cluster_centers[
                    i : i + 1, :
                ] = assigned_points_global / points_in_cluster_global.clamp(
                    1.0, torch.iinfo(torch.int64).max
                )

            # Check whether centroid movement has converged.
            self._inertia = ((self._cluster_centers - new_cluster_centers) ** 2).sum()
            self._cluster_centers = new_cluster_centers.clone()
            if self.tol is not None and self._inertia <= self.tol:
                break
        return self


if __name__ == "__main__":
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size

    if rank == 0:
        print(
            "##############################################\n"
            "# Sample-parallel PyTorch k-means clustering #\n"
            "##############################################"
        )

    data_path = "/pfs/work7/workspace/scratch/ku4408-VL-ScalableAI/data/cityscapes_300.h5"
    dataset = "cityscapes_data"

    if rank == 0:
        print(f"\nLoading data... {data_path}[{dataset}]\n")

    # Data is available in HDF5 format.
    # An HDF5 file is a container for two kinds of objects:
    # - datasets: array-like collections of data
    # - groups: folder-like containers holding datasets and other groups
    # Most fundamental thing to remember when using h5py is:
    # Groups work like dictionaries, and datasets work like NumPy arrays.
    with h5py.File(data_path, "r") as handle:
        chunk = int(
            handle[dataset].shape[0] / size
        )  # Calculate local number of samples in each chunk.
        # Load data chunks from file.
        if rank == size - 1:
            data = torch.tensor(handle[dataset][rank * chunk :], dtype=torch.float)
        else:
            data = torch.tensor(
                handle[dataset][rank * chunk : (rank + 1) * chunk], dtype=torch.float
            )

    print(f"Rank {rank}/{size}: \t[OK]")

    # k-means hyperparameters
    num_clusters = 8
    num_iterations = 20

    kmeans_clusterer = KMeans(
        comm=comm, n_clusters=num_clusters, max_iter=num_iterations
    )
    if rank == 0:
        print("Start fitting the data...")
        start = time.perf_counter()  # Start runtime measurement.

    kmeans_clusterer.fit(data)  # Perform actual k-means clustering.

    if rank == 0:
        print(
            f"DONE.\nRun time:\t{time.perf_counter()-start} s"
        )  # Print measured runtime.
