"""Parallel Sorting by Regular Sampling"""
import argparse
import time

import h5py
import numpy
import torch
from mpi4py import MPI

__mpi_type_mappings = {
    torch.bool: MPI.BOOL,
    torch.uint8: MPI.UNSIGNED_CHAR,
    torch.int8: MPI.SIGNED_CHAR,
    torch.int16: MPI.SHORT,
    torch.int32: MPI.INT,
    torch.int64: MPI.LONG,
    torch.bfloat16: MPI.INT16_T,
    torch.float16: MPI.INT16_T,
    torch.float32: MPI.FLOAT,
    torch.float64: MPI.DOUBLE,
    torch.complex64: MPI.COMPLEX,
    torch.complex128: MPI.DOUBLE_COMPLEX,
}


def sort(a: torch.Tensor, comm: MPI.Comm = MPI.COMM_WORLD) -> torch.Tensor:
    """
    Sort a's elements along given dimension in ascending order by their value.

    The sorting is not stable which means that equal elements in the result may have different ordering than in
    original array.

    Parameters
    ----------
    a : torch.Tensor
        The 1D input array to be sorted.
    comm : MPI.Comm
        The communicator to use.

    Returns
    -------
    torch.Tensor
        Sorted local results.
    """
    size, rank = comm.size, comm.rank

    if size == 1:
        local_sorted, _ = torch.sort(a)
        return local_sorted

    else:
        ###########
        # PHASE 1 #
        ###########
        # P is comm size, n is overall number of samples.
        # Each rank sorts its local chunk and chooses P regular samples as representatives.
        if rank == 0:
            print(
                "###########\n"
                "# PHASE 1 #\n"
                "###########"
            )
        local_sorted, local_indices = torch.sort(a)
        print(f"Rank {rank}/{size}: Local sorting done...[OK]")

        n_local = torch.numel(local_sorted)  # Number of elements in local chunk.
        print(f"Rank {rank}/{size}: Number of elements in local chunk is {n_local}.")
        n_global = comm.allreduce(n_local, op=MPI.SUM)

        # Each rank chooses P regular samples.
        # For this, separate sorted tensor into P+1 equal-length partitions.
        # Regular samples have indices 0, w, 2w,...,(P−1)w where w=n/P^2.
        # Here: `size` = P
        w = int(n_global / size ** 2)
        partitions = [idx * w for idx in range(0, size)]
        regular_samples_local = local_sorted[partitions]
        print(
            f"Rank {rank}/{size}: There are {len(partitions)} local regular samples: {regular_samples_local}"
        )

        # Root gathers regular samples.
        # Each processor has drawn `size` regular samples and we have `size` processors overall.
        num_regular_samples_global = size * size
        if rank == 0:
            print(f"Overall number of regular samples is {num_regular_samples_global}.")
        # Initialize buffer to gather all regular samples on root.
        regular_samples_global = torch.zeros(num_regular_samples_global, dtype=a.dtype)
        comm.Gather(regular_samples_local, regular_samples_global, root=0)
        if rank == 0:
            print("On root: Regular samples gathered...[OK]")

        ###########
        # PHASE 2 #
        ###########
        # Root
        # - sorts gathered regular samples,
        # - chooses P-1 pivots at index positions P + rho, 2P + rho, ..., (P-1)P + rho with rho = P/2 -1, and
        # - shares them with other processes.
        if rank == 0:
            print(
                "###########\n"
                "# PHASE 2 #\n"
                "###########"
            )
        # Initialize buffer for sharing pivots later.
        global_pivots = torch.zeros((size - 1,), dtype=local_sorted.dtype)
        if rank == 0:
            # Sort gathered regular samples on root.
            sorted_regular_samples_global, _ = torch.sort(regular_samples_global)
            print(f"On root: Regular samples are {sorted_regular_samples_global}.")
            # Choose P-1 pivot indices (P = `size`).
            rho = int(size/2) - 1
            global_partitions = [
                idx * size + rho for idx in range(1, size)
            ]
            global_pivots = sorted_regular_samples_global[global_partitions]
            if len(global_partitions) == size - 1:
                print(
                    f"On root: There are {len(global_partitions)} global pivots: {global_pivots}"
                )
        comm.Bcast(
            global_pivots, root=0
        )  # Broadcast copy of pivots to all processes from root.
        if rank == 0:
            print("Pivots broadcast to all processes...")
        ###########
        # PHASE 3 #
        ###########
        if rank == 0:
            print(
                "###########\n"
                "# PHASE 3 #\n"
                "###########\n"
                "Each processor forms P disjunct partitions of locally sorted elements using pivots as splits."
            )
        # Each processor forms P disjunct partitions of locally sorted elements using P-1 pivots as splits.
        # Initialize buffer to store partition mask of local samples.
        # `lt_partitions` is a tensor of shape [P, n_local]. We use it to assign local samples to the partitions as
        # split by the global pivots. This is done via a boolean mask.
        lt_partitions = torch.empty((size, local_sorted.shape[0]), dtype=torch.int64)
        last = torch.zeros_like(local_sorted, dtype=torch.int64)
        # Iterate over all pivots and store index of first pivot greater than element's value.
        if rank == 0:
            print("Iterate over pivots to find index of first pivot > element's value.")

        for idx, pivot in enumerate(global_pivots):
            # torch.lt(input, other, *, out=None) computes `input < other` element-wise.
            # Returns boolean tensor that is True where input is less than other and False elsewhere.
            lt = torch.lt(local_sorted, pivot).int()
            lt_partitions[idx] = lt - last
            last = lt

        lt_partitions[size - 1] = torch.ones_like(local_sorted, dtype=last.dtype) - last

        # `lt_partitions` contains P elements: the first one encodes which elements in `local_sorted` are smaller than
        # the 1st (= smallest) pivot, the second one encodes which elements are larger than the 1st and smaller than the
        # 2nd pivot, ..., and the last encodes which elements are larger than last ( = largest) pivot.
        # Now each processor has partitioned its locally sorted data into P partitions.
        # To obtain the globally sorted result, each processor needs to send its i-th partition to processor i so that
        # each processor i finally holds all samples from partition i.
        # This all-to-all communication requires information about how many values will be sent for each partition.
        # Processor i keeps i-th partitions and sends j-th partition to processor j.
        local_partition_sizes = torch.sum(
            lt_partitions, dim=1
        )  # How many values are in each local partition?
        print(
            f"Rank {rank}/{size}: Local # elements to be sent to other ranks (keep own section): {local_partition_sizes}"
        )
        global_partition_sizes = torch.zeros_like(
            local_partition_sizes
        )  # How many values are in each global partition?
        comm.Allreduce(local_partition_sizes, global_partition_sizes, op=MPI.SUM)
        if rank == 0:
            print(
                f"Global # of elements on all ranks (partition matrix): {global_partition_sizes}"
            )
        # Construct matrix holding information which local value will be shipped where.
        communication_index_vector = torch.empty_like(local_sorted, dtype=torch.int64)
        # Loop over `lt_partitions` (binary encoding of which elements is in which partition formed by pivots).
        for idx_rank, partition_mask_value in enumerate(lt_partitions):
            communication_index_vector[partition_mask_value > 0] = idx_rank
            # Elements in 0th partition (< first pivot) get 0, i.e., will be collected at rank 0, elements in 1st
            # partition (> than first + < than second pivot) get 1, i.e., will be collected at rank 1,...
        print(f"Rank {rank}/{size}: Ship element to rank: {communication_index_vector}")
        # Counts and displacements for Alltoallv are rank-specific!
        # send_counts_local on rank i: Integer array, entry j specifies number of values to be sent to rank j.
        # recv_counts_local on rank i: Integer array, entry j specifies number of values to be received from rank j.
        # Determine how many elements are to be shipped to which rank.
        send_counts_local = numpy.zeros(size, dtype=int)
        # Determine number of elements in each local partition.
        for idx_rank in numpy.arange(size):
            send_counts_local[idx_rank] = int((communication_index_vector == idx_rank).sum(dim=0))
        # Determine local send displacements from local send counts.
        send_displ_local = numpy.zeros(size, dtype=int)
        send_displ_local[1:] = numpy.cumsum(send_counts_local, axis=0)[:-1]
        # Allgather local send counts to determine how many elements are sent by other ranks to this rank and thus are
        # to be received.
        send_counts_global = numpy.zeros((size, size), dtype=int)
        comm.Allgather([send_counts_local, MPI.INT], [send_counts_global, MPI.INT])
        # Determine global receive counts from transposed global send counts.
        recv_counts_global = numpy.transpose(send_counts_global)
        # Extract local receive counts.
        recv_counts_local = recv_counts_global[rank]
        # Determine local receive displacements from local receive counts.
        recv_displ_local = numpy.zeros(size, dtype=int)
        recv_displ_local[1:] = numpy.cumsum(recv_counts_local, axis=0)[:-1]

        value_buffer = torch.zeros((global_partition_sizes[rank],), dtype=local_sorted.dtype)
        # Communicating torch tensors using buffered MPI variants requires accessing the tensor's storage pointer!
        send_buf = [
            MPI.memory.fromaddress(local_sorted.data_ptr(), 0),
            (send_counts_local.tolist(), send_displ_local.tolist()),
            __mpi_type_mappings[local_sorted.dtype],
        ]
        recv_buf = [
            MPI.memory.fromaddress(value_buffer.data_ptr(), 0),
            (recv_counts_local.tolist(), recv_displ_local.tolist()),
            __mpi_type_mappings[value_buffer.dtype],
        ]
        comm.Alltoallv(send_buf, recv_buf)
        result, _ = torch.sort(value_buffer)
        return result


if __name__ == "__main__":
    data_path = "/pfs/work7/workspace/scratch/ku4408-VL-ScalableAI/data/psrs_data.h5"

    parser = argparse.ArgumentParser(prog="Parallel Sorting by Regular Samples")
    parser.add_argument(
        "--dataset",
        type=str,
        default="duplicates_1",
        help="The dataset to be sorted.",
    )

    args = parser.parse_args()

    comm = MPI.COMM_WORLD
    rank, size = comm.rank, comm.size

    with h5py.File(data_path, "r") as f:
        chunk = int(f[args.dataset].shape[0] / size)
        if rank == size - 1:
            data = torch.tensor(f[args.dataset][rank * chunk:])
        else:
            data = torch.tensor(f[args.dataset][rank * chunk:(rank + 1) * chunk])

    if rank == 0:
        print(
            "########\n"
            "# PSRS #\n"
            "########"
        )

    print(f"Local data on rank {rank} = {data}")

    if rank == 0:
        print("Start sorting...")
    start = time.perf_counter()
    result = sort(data)
    elapsed_local = time.perf_counter() - start
    elapsed_global = comm.allreduce(elapsed_local, op=MPI.SUM)
    elapsed_global /= size
    if rank == 0:
        print(f"Sorting done...\nRank-averaged run time: {elapsed_global} s")
    print(f"Sorted chunk on rank {rank}/{size}: {result}")
